import crossnetwork_processing_functions as cpf
import datetime
import GHOST_standards as Gs
import numpy as np
import standardisation_mapping_functions as smf
import time
import os

# create dictionary that map reading parameter to attribute name of ISDLineParser
mapar = {'WND_DIR': 'direction_angle', 'WND_DIR_10M': 'direction_angle', 'WND_SPD': 'speed',
         'WND_SPD_10M': 'speed', 'CEILING_HEIGHT': 'ceiling', 'VIS_DIST': 'distance',
         'TEMP': 'air_temp', 'TEMP_2M': 'air_temp', 'DEW_PT': 'dpoint_temp', 'DEW_PT_2M': 'dpoint_temp',
         'SLP': 'sea_level_pressure', 'PREC_ACCUM': 'prec_depth', 'SNOW_ACCUM': 'snow_acc',
         'SNOW_DEPTH': 'si', 'PRES': 'station_p', 'PRES_2M': 'station_p', 'PRES_10M': 'station_p',
         'SST': 'sst', 'SOIL_TEMP': 'soil_temp', 'SOIL_TEMP_10CM': 'soil_temp', 'SOIL_TEMP_40CM': 'soil_temp',
         'SOIL_TEMP_100CM': 'soil_temp', 'SOIL_TEMP_200CM': 'soil_temp', 'CLOUD_CVG': 'cloud',
         'CLOUD_CVG_FRAC': 'cloud_frac', '`RH': 'avg_rh'}


def read_data(slurm_job_ID, parameter, station_reference):
    """Convert data provider's original file for a specific intro a GHOST standard .npy station specific file"""

    # initialise log file string
    log_file_str = ''

    log_file_str += '{} -- PARSING RAW DATA\n'.format(station_reference)

    # Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = cpf.get_run_dictionaries(slurm_job_ID, 'NCDC_ISD', parameter)
    # time start
    run_details['stage_process_start_time'] = time.time()

    # get data files specific to station
    data_files = cpf.get_data_files(run_details, parameter_details, specific_station=station_reference)

    # sort files in date order
    data_file_start_dates = [int(data_file.split('-')[-1]) for data_file in data_files]
    data_files = np.array([x for _, x in sorted(zip(data_file_start_dates, data_files))])

    # set flag to know when have written data for station
    written_data = False

    # get parameter's attribute name of ISDLineParser
    reading = mapar[parameter]

    # get dictionary necessary for mapping of NCDC file data flags to standard data flag codes
    data_flag_map = smf.network_data_flag_to_standard_data_flag_mapping['NCDC_ISD']

    # get metadata file
    metafile = np.load(run_details['data_root'] + run_details['network'] + '/metadata/processed/' + run_details[
        'network'] + '.npy')

    # Iteration over the files starts here
    # iterate through site files
    for file_n, data_file in enumerate(data_files):

        log_file_str += 'PROCESSING FILE: {}\n'.format(data_file)

        # find station's line in the metadata file
        name = data_file[-17:]
        metaline = metafile[(metafile['USAF'] == name[:6]) & (metafile['WBAN'] == name[7:12])]
        # if there is no entry for file in metadata, discard file
        if len(metaline) == 0:
            log_file_str += "{} does not have metadata entry, skipping\n".format(data_file)
            break
        station_elv = metaline['ELEV(M)'][0]
        if station_elv != '':
            station_elv = round(float(metaline['ELEV(M)'][0]))
        else:
            log_file_str += "no metadata elevation for {}, skipping".format(data_file)
            break

        # create site instances of metadata/data dictionaries
        site_data = Gs.get_standard_data(parameter_details)
        site_metadata = Gs.get_standard_metadata(parameter_details)

        # initialize list that will contain all objects/lines of file
        lines = []  # contains final lines we process
        data_lines = []  # list of lines of the entire file, from which we'll not keep everything

        # first step for NCDC_ISD: correct measurement altitudes and sampling heights
        unique_data_elevations = []
        unique_samplings = []
        with open(data_file, 'r') as f:
            for data_line in f.readlines():
                data_line = cpf.ISDLineParser(data_file, data_line)
                data_line.station_elevation = station_elv
                if data_line.elevation is np.NaN:
                    continue
                data_lines.append(data_line)
                if data_line.elevation not in unique_data_elevations:
                    # keep the unique data elevations and sampling heights
                    unique_data_elevations.append(data_line.elevation)
                    unique_samplings.append(data_line.elevation - station_elv)

        # keep indexes where the sampling height is non-negative
        positive_samp = [idx for idx, x in enumerate(unique_samplings) if x >= 0]
        # initialize flag, whether all the absolute difference are >= 10m
        flag = False

        for data_line in data_lines:
            sampling = data_line.elevation - station_elv  # take the sampling height
            if not all(values < 0 for values in unique_samplings):
                # initialize list for differences
                diffs = []
                if sampling >= 0:
                    # we assume that it is correct, so we keep the initial values
                    data_line.sampling_height = sampling
                    data_line.measurement_altitude = data_line.elevation
                elif sampling < 0:
                    # iterate through the unique elevations whose associated sampling height is non-negative
                    for idx in positive_samp:
                        abs_diff = abs(data_line.elevation - unique_data_elevations[idx])
                        diffs.append(abs_diff)  # append here all the differences
                    if all(d >= 10 for d in diffs):
                        flag = True
                    else:
                        closest = min([d for d in diffs if d < 10])
                        idx = diffs.index(closest)
                        data_line.measurement_altitude = unique_data_elevations[positive_samp[idx]]
                        data_line.sampling_height = data_line.measurement_altitude - station_elv

            if all(values < 0 for values in unique_samplings) or flag:
                new_unique_data_elevations = unique_data_elevations.copy()
                new_unique_data_elevations.append(station_elv)
                new_unique_samplings = unique_samplings.copy()
                new_unique_samplings.append(0.0)
                new_positive_samp = [idx for idx, x in enumerate(new_unique_samplings) if x >= 0]
                if sampling >= 0:
                    # we assume that it is correct, so we keep the initial values
                    data_line.sampling_height = sampling
                    data_line.measurement_altitude = data_line.elevation
                elif sampling < 0:
                    diffs = []
                    if len(new_positive_samp) > 0:
                        for idx in new_positive_samp:
                            abs_diff = abs(data_line.elevation - new_unique_data_elevations[idx])
                            diffs.append(abs_diff)
                    if all(values >= 50 for values in diffs):
                        data_line.sampling_height = data_line.elevation
                        data_line.measurement_altitude = data_line.sampling_height + station_elv
                    else:
                        closest = min([d for d in diffs if d < 50])
                        idx = diffs.index(closest)
                        data_line.measurement_altitude = new_unique_data_elevations[new_positive_samp[idx]]
                        data_line.sampling_height = data_line.measurement_altitude - station_elv

        # by now, all data lines should have the corrected sample height and measurement altitude

        # from all the data_lines that contain all everything from the file, keep only those data_lines
        # that contain the parameter that we are reading
        for data_line in data_lines:
            if 'SOIL_TEMP_' in parameter:
                if parameter == 'SOIL_TEMP_10CM' and hasattr(data_line, reading):
                    if data_line.soil_temp_depth == 10:
                        lines.append(data_line)
                    if parameter == 'SOIL_TEMP_40CM' and hasattr(line, reading):
                        if data_line.soil_temp_depth == 40:
                            lines.append(data_line)
                    if parameter == 'SOIL_TEMP_100CM' and hasattr(line, reading):
                        if data_line.soil_temp_depth == 100:
                            lines.append(data_line)
                    if parameter == 'SOIL_TEMP_200CM' and hasattr(line, reading):
                        if data_line.soil_temp_depth == 200:
                            lines.append(data_line)
            elif '2M' in parameter:
                if hasattr(data_line, reading) and (data_line.sampling_height <= 4) and (
                        data_line.sampling_height >= 0):
                    data_line.sampling_height = 2.0
                    data_line.measurement_altitude = data_line.station_elevation + 2.0
                    lines.append(data_line)
                elif '10M' in parameter:
                    if hasattr(data_line, reading) and (data_line.sampling_height <= 12) and (
                            data_line.sampling_height >= 8):
                        data_line.sampling_height = 10.0
                        data_line.measurement_altitude = data_line.station_elevation + 10.0
                        lines.append(data_line)
                else:
                    if hasattr(data_line, reading):
                        lines.append(data_line)

        # if length of s_times == 0, then do not process file, as zero valid data
        if len(lines) == 0:
            log_file_str += 'NO DATA IN FILE: {}. SKIPPING.'.format(data_file)
            break

        # ----------------------------------------------------------------------------#
        # PROCESS METADATA
        # ----------------------------------------------------------------------------#

        # initialize dictionary to keep unique lat and lon as value, and where they appear as key
        uniqll = {}

        # append metadata as many times as the parameter being read is present
        # (lines list contains only the objects/lines of the file that contain the
        # parameter we are currently working on)
        i = 0  # keep a counter needed for the indexes to populate station timezone
        for line in lines:
            # find station's line
            metaline = metafile[(metafile['USAF'] == line.station_idu) & (metafile['WBAN'] == line.station_idn)]
            # if station doesn't exist in metadata file
            if len(metaline) == 0:
                cntry, sname, prov = ('' for _ in range(3))
            else:
                cntry = metaline['CTRY'][0]
                sname = metaline['STATION NAME'][0]
                prov = metaline['STATE'][0]

            # GHOST version
            site_metadata['GHOST_version']['value'].append(run_details['ghost_version'])
            # data_timezone, 0 for all as time is reported in UTC
            site_metadata['data_timezone']['value'].append(0)
            # latitude and longitude: check if coordinates are suspect
            new_longitude, new_latitude = cpf.check_geographical_position(line.longitude, line.latitude)
            # and append
            site_metadata['longitude']['value'].append(new_longitude)
            site_metadata['latitude']['value'].append(new_latitude)

            # altitude
            if line.station_elevation == '':
                site_metadata['altitude']['value'].append(np.NaN)
            else:
                site_metadata['altitude']['value'].append(line.station_elevation)
            # measurement altitude
            site_metadata['measurement_altitude']['value'].append(line.measurement_altitude)
            # country
            if cntry == '':
                site_metadata['country']['value'].append(np.NaN)
            else:
                if cntry in Gs.country_FIPS104_to_name:
                    site_metadata['country']['value'].append(Gs.country_FIPS104_to_name[cntry])
                # if they aren't consistent with the country encodings, use ISO
                elif cntry in Gs.country_ISO3166_to_name:
                    site_metadata['country']['value'].append(Gs.country_ISO3166_to_name[cntry])
                # otherwise, we can't do anything, we tried
                else:
                    site_metadata['country']['value'].append(np.NaN)
            # station_name
            if sname == '':
                site_metadata['station_name']['value'].append(np.NaN)
            else:
                site_metadata['station_name']['value'].append(sname)
            # province
            if prov == '':
                site_metadata['province']['value'].append(np.NaN)
            else:
                site_metadata['province']['value'].append(prov)

            # append network's name
            site_metadata['network']['value'].append('NCDC_ISD')

            # process and add associated networks.
            # Format of field is -> USAF: 00724; WBA: 9182; ICAO:ENRS
            anet = []
            if line.station_idu != '999999':
                usaf = 'USAF:' + line.station_idu
                anet.append(usaf)
            if line.station_idn != '99999':
                wban = 'WBAN:' + line.station_idn
                anet.append(wban)
            if metafile['ICAO'][0] != '':
                icao = 'ICAO' + metafile['ICAO'][0]
                anet.append(icao)
            if len(anet) != 0:
                res = ''
                for net in anet:
                    res = res + net + ';'
                res = res[:-1]
                site_metadata['associated_networks']['value'].append(res)

            # metadata update stamp
            year = int(line.datetime_ascii[0:4])
            month = int(line.datetime_ascii[4:6])
            day = int(line.datetime_ascii[6:8])
            hour = int(line.datetime_ascii[8:10])
            minute = int(line.datetime_ascii[10:12])
            date = cpf.datetime_to_fractional_minutes_from_reference(datetime.datetime(year, month, day, hour, minute))
            # and finally append to the dictionary
            site_metadata['meta_update_stamp']['value'].append(date)

            # contact institution
            site_metadata['contact_institution']['value'].append("NCDC_ISD")
            # contact email
            site_metadata['contact_email_address']['value'].append("ncei.orders@noaa.gov")

            # finally, keep the unique lats and lons and their indexes
            pair = (line.latitude, line.longitude)
            if pair in uniqll:
                uniqll[pair].append(i)
            else:
                uniqll[pair] = []
                uniqll[pair].append(i)
            i += 1  # update counter

            # ----------------------------------------------------------------------------#
            # PROCESS DATA
            # ----------------------------------------------------------------------------#

            # initialize lists for flags and qa
            flags = []
            qa = []

            # time of measurement, according to ghost standard
            site_data['time']['value'].append(date)
            # get variable's measurement and append it
            meas = getattr(line, reading)
            site_data[parameter_details['bsc_parameter_name']]['value'].append(meas)
            # all the variables are instantaneous, except rain and snow, which are treated separately
            # assign 0 to the rest
            if parameter == "PREC_ACCUM":
                site_data['temporal_resolution']['value'].append(line.prec_period)
            elif parameter == "SNOW_ACCUM":
                site_data['temporal_resolution']['value'].append(line.snow_acc_period)
            elif parameter == "AVG_RH":
                site_data['temporal_resolution']['value'].append(line.avg_rh_period)
            else:
                site_data['temporal_resolution']['value'].append(0) 

            # flags
            # same for variable's measurement flag
            meas_qc = getattr(line, reading + '_qc')
            # get standard name of needed to get standard code
            stflag = data_flag_map[meas_qc]['standard_data_flag_name']
            # then use name to get the flag's code from Gs dictionary
            stand_code = Gs.standard_data_flag_name_to_data_flag_code[stflag]
            # and here is the code, now we can append it
            flags.append(stand_code)
            site_data['flag']['value'].append(flags)

            # qa
            ghost_decreed_validity = data_flag_map[meas_qc]['GHOST_decreed_validity']
            network_decreed_validity = data_flag_map[meas_qc]['network_decreed_validity']
            # append qa flag code if standard data flag is GHOST decreed invalid
            if ghost_decreed_validity == 'I':
                qa.append(Gs.standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
            # append qa flag if standard data flag is network decreed invalid
            if network_decreed_validity == 'I':
                qa.append(Gs.standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
            # and finally, append the qa code list for this measure to the dict field
            site_data['qa']['value'].append(qa)

            # finally, the rest of the fields do not have data from NCDC_ISD
            # append an empty list per measuremet for each of the fields
            # lower limit, should append NaN, not empty list
            site_data['reported_lower_limit_of_detection_per_measurement']['value'].append(np.NaN)
            # upper limit, should append NaN, not empty list
            site_data['reported_upper_limit_of_detection_per_measurement']['value'].append(np.NaN)
            # reported uncertainty per measurement, should append NaN, not empty list
            site_data['reported_uncertainty_per_measurement']['value'].append(np.NaN)
            # derived uncertainty per measurement, should append NaN, not empty list
            site_data['derived_uncertainty_per_measurement']['value'].append(np.NaN)
            # classification, here should append empty list, not NaN
            site_data['classification']['value'].append([])

        if len(lines) != 0:

            # save data_download_stamp --> rounded to nearest minute
            data_download_datetime = datetime.datetime.strptime(time.ctime(os.path.getmtime(data_file)),
                                                                "%a %b %d %H:%M:%S %Y")
            data_download_datetime = cpf.round_to_minute(data_download_datetime)
            # create temp list to append the result
            tmp = [cpf.datetime_to_fractional_minutes_from_reference(data_download_datetime)]
            # populate dict's fild with the value as many as required
            site_metadata['data_download_stamp']['value'] = tmp * len(site_metadata['latitude']['value'])

            site_metadata, measurement_qa_flags, standard_method_abbreviation = smf.\
                map_network_measurement_information_to_standard_measurement_information('', site_metadata, run_details)

            # append station_reference now that we have the standard method abbreviation
            site_metadata['station_reference']['value'] = \
                ['{}_{}'.format(station_reference, standard_method_abbreviation)] \
                * len(site_metadata['latitude']['value'])

            for standard_flag_name in list(measurement_qa_flags.keys()):
                if measurement_qa_flags[standard_flag_name]:
                    for qq in range(len(site_data['qa']['value'])):
                        site_data['qa']['value'][qq].append(Gs.standard_QA_name_to_QA_code[standard_flag_name])

            # treat timezone
            # fist, populate the list with NaNs, as otherwise we cannot fill specific positions
            site_metadata['station_timezone']['value'] = [np.NaN] * len(site_metadata['latitude']['value'])
            # then take date reference
            dt_ref = datetime.datetime(2000, 1, 1, 0, 0)
            # and finally populate the timezone value in the positions that the lat-lon appear
            for key in uniqll.keys():
                lat, lon = key
                pytz_str, timezn = cpf.local_offset_from_utc(dt_ref, latitude=lat, longitude=lon)
                for pos in uniqll[key]:
                    site_metadata['station_timezone']['value'][pos] = timezn

            # iterate through metadata variables, if any columns have not been appended to add a NaN
            # check number of refs in list
            for column_header in list(site_metadata.keys()):
                n_column = len(site_metadata[column_header]['value'])
                # if column is empty, then append NaN
                if n_column < 1:
                    site_metadata[column_header]['value'] = [np.NaN] * len(site_metadata['latitude']['value'])

            # standardise metadata/data and write out specific file station data to temporary .npy
            cpf.post_process_file_parse(site_metadata, site_data, run_details, parameter_details)

            # update flag to state that have written data for station
            written_data = True

    # if have written data for station, return valid process code: 0
    if written_data:
        log_file_str += str((time.time() - run_details['stage_process_start_time']) / 60.)
        cpf.create_GHOST_stage_logfile('S1', run_details, 0, log_file_str, station_reference)
    # otherwise no data has been written for station, therefore process code to state station will
    # not continue in further processing (2)
    else:
        log_file_str += 'NO DATA HAS BEEN WRITTEN FOR STATION'
        cpf.create_GHOST_stage_logfile('S1', run_details, 1, log_file_str, station_reference)
